"""Imports for the text parser."""

from plaso.parsers.text_plugins import android_logcat
from plaso.parsers.text_plugins import apache_access
from plaso.parsers.text_plugins import apt_history
from plaso.parsers.text_plugins import atlassian_bitbucket
from plaso.parsers.text_plugins import atlassian_confluence
from plaso.parsers.text_plugins import atlassian_jira
from plaso.parsers.text_plugins import aws_elb_access
from plaso.parsers.text_plugins import bash_history
from plaso.parsers.text_plugins import bitbucket_access
from plaso.parsers.text_plugins import bitbucket_audit
from plaso.parsers.text_plugins import confluence_access
from plaso.parsers.text_plugins import cri
from plaso.parsers.text_plugins import dpkg
from plaso.parsers.text_plugins import gdrive_synclog
from plaso.parsers.text_plugins import google_logging
from plaso.parsers.text_plugins import iis
from plaso.parsers.text_plugins import ios_lockdownd
from plaso.parsers.text_plugins import ios_logd
from plaso.parsers.text_plugins import ios_sysdiag_log
from plaso.parsers.text_plugins import jira_access
from plaso.parsers.text_plugins import macos_appfirewall
from plaso.parsers.text_plugins import macos_launchd
from plaso.parsers.text_plugins import macos_securityd
from plaso.parsers.text_plugins import macos_wifi
from plaso.parsers.text_plugins import popcontest
from plaso.parsers.text_plugins import postgresql
from plaso.parsers.text_plugins import powershell_transcript
from plaso.parsers.text_plugins import santa
from plaso.parsers.text_plugins import sccm
from plaso.parsers.text_plugins import selinux
from plaso.parsers.text_plugins import setupapi
from plaso.parsers.text_plugins import skydrivelog
from plaso.parsers.text_plugins import snort_fastlog
from plaso.parsers.text_plugins import sophos_av
from plaso.parsers.text_plugins import syslog
from plaso.parsers.text_plugins import teamviewer
from plaso.parsers.text_plugins import viminfo
from plaso.parsers.text_plugins import vsftpd
from plaso.parsers.text_plugins import winfirewall
from plaso.parsers.text_plugins import xchatlog
from plaso.parsers.text_plugins import xchatscrollback
from plaso.parsers.text_plugins import zsh_extended_history
