"""Imports for the JSON-L parser plugins."""

from plaso.parsers.jsonl_plugins import aws_cloudtrail_log
from plaso.parsers.jsonl_plugins import azure_activity_log
from plaso.parsers.jsonl_plugins import azure_application_gateway_log
from plaso.parsers.jsonl_plugins import docker_container_config
from plaso.parsers.jsonl_plugins import docker_container_log
from plaso.parsers.jsonl_plugins import docker_layer_config
from plaso.parsers.jsonl_plugins import gcp_log
from plaso.parsers.jsonl_plugins import ios_app_privacy
from plaso.parsers.jsonl_plugins import microsoft365_audit_log
